{{! narwhals docstring template }}
{{summaryPlaceholder}}.
{{#parametersExist}}

Arguments:
{{#args}}
    {{var}}: {{descriptionPlaceholder}}
{{/args}}
{{#kwargs}}
    {{var}}: {{descriptionPlaceholder}}
{{/kwargs}}
{{/parametersExist}}
